
IMPLEMENTATION MODULE Rally;
IMPORT IO;
IMPORT FIO;

IMPORT Lib;
FROM Storage IMPORT ALLOCATE, DEALLOCATE, Available;

TYPE
   RALLY=POINTER TO TABLA;
   LISTA=POINTER TO NODO;
   TABLA=ARRAY ['A'..'Z'] OF LISTA;

   NODO=RECORD
             nombre:MiStr.TCadena;
             numero, velocidad:INTEGER;
             crono:LONGCARD;(* Tiempo en segundos de la ultima actualizacion.*)
             vueltas,metros:CARDINAL; (* Vueltas recorridas y metros sueltos *)
             sig:LISTA;
        END;

PROCEDURE Crear(VAR bien:BOOLEAN):RALLY;
VAR
  r:RALLY;
  i:CHAR;
BEGIN
    bien:=TRUE;
    IF Available(SIZE(TABLA)) THEN
       NEW(r);
       FOR i:='A' TO 'Z' DO
           r^[i]:=NIL;
       END;
     ELSE
        bien:=FALSE;     (* Error de sin memoria *)
        r:=NIL;
     END;
     RETURN r;
END Crear;

PROCEDURE EsVacio(r:RALLY;VAR bien:BOOLEAN):BOOLEAN;
VAR
  letra:CHAR;
  ret:BOOLEAN;
BEGIN
    ret:=FALSE;
    bien:=TRUE;
    IF r=NIL THEN
       bien:=FALSE;    (* error de estrutura no creada *)
    ELSE
       letra:='A';
       WHILE (letra<='Z') AND (r^[letra]=NIL) DO
             letra:=CHR(ORD(letra)+1);
       END;
       ret:=(letra=CHR(ORD('Z')+1));
    END;
    RETURN ret;    (*Devuelve FALSE si hubo un error o no esta vacio*)
END EsVacio;


(* Este procedimiento devuelve TRUE si la escuderia ya esta en el rally *)
PROCEDURE EstaEsc(l:LISTA;e:ARRAY OF CHAR):BOOLEAN;

VAR
aux:LISTA;
BEGIN
    aux:=l;
    WHILE (aux<>NIL) AND (MiStr.Comparar(e,aux^.nombre)<>0) DO
          aux:=aux^.sig;
    END;
    RETURN aux<>NIL;
END EstaEsc;


(* Este procedimiento devuelve TRUE si el n de coche esta en el rally *)
PROCEDURE EstaNum(r:RALLY; n:INTEGER):BOOLEAN;

VAR
  aux:LISTA;
  p:CHAR;
  esta:BOOLEAN;
BEGIN
    esta:=FALSE;
    p:='A';
    WHILE (NOT(esta)) AND (p<='Z') DO
         aux:=r^[p];
         WHILE (aux<>NIL) AND (esta=FALSE) DO
               IF aux^.numero=n THEN
                  esta:=TRUE;
               END;
               aux:=aux^.sig;
         END;
         p:=CHR(ORD(p)+1);
    END;
    RETURN esta;
END EstaNum;

(*  Este procedimiento se encarga de colocar un coche n
    en la estructura; es llamado por el procedimiento introducir.*)
PROCEDURE Colocar(VAR r:RALLY; n:LISTA);

VAR
  aux, anterior:LISTA;
  p:CHAR;
BEGIN
    p:=CAP(n^.nombre[0]); (* Inicial de la escuderia *)
    IF r^[p]=NIL THEN     (* Si la lista que tiene por indice la *)
       r^[p]:=n;          (* de la escuderia esta vacia =>       *)
    ELSE                  (*  insertamos en la primera posicion. *)
       aux:=r^[p];        (* En otro caso...*)
       IF (MiStr.Comparar(aux^.nombre,n^.nombre)=1) THEN
          (* Si no esta vacia pero debemos insertar en la 1 posicion.*)
          n^.sig:=aux;
          r^[p]:=n;
       ELSE
          WHILE (aux<>NIL) AND (MiStr.Comparar(aux^.nombre,n^.nombre)<0) DO
                anterior:=aux;
                aux:=aux^.sig;
          END;
          n^.sig:=aux;
          anterior^.sig:=n;
       END;
    END;
END Colocar;


(* Este procedimiento devuelve el tiempo actual en segundos *)
PROCEDURE Tiempo():LONGCARD;
VAR
  h, m, s, d:CARDINAL;        (* horas,minutos,segundos,decimas de segundo*)
  total, ht, mt:LONGCARD;
BEGIN
    Lib.GetTime(h,m,s,d);
    ht:=LONGCARD(h)*3600;
    mt:=LONGCARD(m)*60;
    total:=ht+mt+LONGCARD(s);
    RETURN total;
END Tiempo;


(* Introduce  un coche en la estructura primero por su inicial y
   posteriormente de forma alfabetica, comprobando a su vez la
   correccion de los datos leidos por teclado:
             velocidad positiva,
             numero positivo,
             solo un coche por escuderia,
             numeros no repetidos.      *)
PROCEDURE Introducir(VAR rally:RALLY; escuderia:ARRAY OF CHAR;
                         numero, velocidad:INTEGER;VAR bien:BOOLEAN);

VAR
  nuevo:LISTA;

BEGIN
    bien:=TRUE;
    IF EstaEsc(rally^[CAP(escuderia[0])],escuderia) THEN
       bien:=FALSE      (* error escuderia repetida *)
    ELSIF EstaNum(rally,numero) THEN
       bien:=FALSE      (* error numero repetido*)
    ELSIF NOT(Available(SIZE(NODO))) THEN
       bien:=FALSE      (* error Sin Memoria *)
    ELSE
       NEW(nuevo);
       MiStr.Copiar(nuevo^.nombre,escuderia);
       nuevo^.numero:=numero;
       nuevo^.velocidad:=velocidad;
       nuevo^.crono:=0;                (*LONGCARD(Tiempo());MMMMM*)
       nuevo^.vueltas:=0;
       nuevo^.metros:=0;
       nuevo^.sig:=NIL;
       Colocar(rally,nuevo);
    END;
END Introducir;

(* Este proc. actualiza el numero de vueltas que recorrio un coche y su crono *)
PROCEDURE  Calcular(VAR l:LISTA);
VAR
    v:CARDINAL;                 (* ta ->tiempo actual, v->vueltas, m->metros *)
    ta, t, m:LONGCARD;
BEGIN
    ta:=Tiempo();
    IF l^.crono<> 0 THEN
       t:=ta-l^.crono;                (* t segundos => Distancia recorrida?*)
       l^.crono:=ta;
       m:=t*LONGCARD(l^.velocidad);   (* metros:= tiempo X velocidad *)
       v:=CARDINAL(m DIV 1000);       (* Vueltas enteras recorridas *)
       m:=(m MOD 1000);               (* Metros sueltos recorridos, a los cuales *)
       m:=m+LONGCARD(l^.metros);       (* les sumo los metros que ya habia de antes *)
       v:=v+CARDINAL(m DIV 1000);     (* y miro si en total suman mas de 1000 metros. *)
       m:=m MOD 1000;
       l^.vueltas:=l^.vueltas+v;      (*Le sumamos los datos calculados al coche*)
       l^.metros:=CARDINAL(m)
     ELSE
       l^.crono:=ta;                   (*banderazo de salida para el coche *)
     END;
END Calcular;


(* Este proc. actualiza las vueltas recorridas por cada uno de los coches *)
PROCEDURE Cronometrar(VAR r:RALLY);
VAR
  l:LISTA;
  p:CHAR;
BEGIN
    p:='A';
    WHILE p<='Z' DO
          l:=r^[p];
          WHILE l<>NIL DO
                Calcular(l);
                l:=l^.sig;
          END;
          p:=CHR(ORD(p)+1);
    END;
END Cronometrar;

PROCEDURE VerCoche(l:LISTA);
BEGIN
  IO.WrLn;
  IO.WrStr("---------------------------------------");
  IO.WrLn;
  IO.WrStr(" Escuderia: ");
  IO.WrStr(l^.nombre);
  IO.WrLn;
  IO.WrStr(" Numero: ");
  IO.WrInt(l^.numero,0);
  IO.WrLn;
  IO.WrStr(" Velocidad media: ");
  IO.WrInt(l^.velocidad,0);
  IO.WrLn;
  IO.WrStr(" Vueltas: ");
  IO.WrCard(l^.vueltas,0);
  IO.WrStr(" y ");
  IO.WrCard(l^.metros,0);
  IO.WrStr(" metros.");
  IO.WrLn;
END VerCoche;


(* Este proc. informa de quien va primero en la clasificacion de la competicion
   usando la velocidad media y el tiempo transcurrido desde la ultima
   medicion, asi como la distancia recorrida hasta entonces. Esta
   distancia se mide en vueltas(1 vuelta son 1000 metros).
   La primera vez que se llama actualiza el campo crono para cada coche con
   el tiempo actual dando con ello el 'banderazo' de salida en la carrera.
       Si un coche supera las 100 vueltas sera el ganador absoluto
   pudiendose dar la posibilidad de empate entre varios coches. *)

PROCEDURE Ganador(VAR rally:RALLY;VAR bien:BOOLEAN);

VAR
  vueltas, metros:CARDINAL;
  letra:CHAR;
  coche:LISTA;
  vacio:BOOLEAN;
BEGIN
  bien:=TRUE;
  vacio:=EsVacio(rally,bien);
  IF vacio THEN     (* TAD vacio (no confundir con no creado*)
    bien:=FALSE     (* No puedo calcular el ganadr de un rally vacio *)
  ELSIF bien THEN   (* Error: el rally no esta vacio.
                     En otro caso devuelve error  *)
    Cronometrar(rally);      (* Actualiza vueltas recorridas por cada coche. *)
    vueltas:=0;              (* Mayor numero de vueltas recorridas. *)
    metros:=0;               (* Metros sueltos *)
    letra:='A';
    WHILE (letra<='Z') DO
                         (* Recorro la estructura buscando el > n de vueltas*)
      coche:=rally^[letra];
      WHILE (coche<>NIL)DO
         IF vueltas<coche^.vueltas THEN (* Si encuentro un valor de vueltas
                                           mayor que el actual => copio tanto
                                           el n de vueltas como de metros. *)
            vueltas:=coche^.vueltas;
            metros:=coche^.metros;
         ELSIF (vueltas=coche^.vueltas) AND (metros<coche^.metros) THEN
                                    (* Si encuentro un valor de vueltas
                                       igual que el actual => miro si el n
                                       de metros es mayor y lo copio si es
                                       mayor. *)
            metros:=coche^.metros;
         END;(*if*)
         coche:=coche^.sig;
       END;(*while*)
       letra:=CHR(ORD(letra)+1);
    END;(*while*)
    IF vueltas>=100 THEN   (* Si un corredor dio mas de 100 vueltas...*)
        IO.WrStr(" El ganador absoluto es : ");
        IO.WrLn;
    END;          (*if*)
    letra:='A';
           (* Recorro de nuevo la estructura y visualizo los participantes
             cuyo numero de vueltas y metros coincida con los obtenidos.*)
    WHILE (letra<='Z') DO
         coche:=rally^[letra];
         WHILE coche<>NIL DO
               IF (vueltas=coche^.vueltas) AND (metros=coche^.metros) THEN
                  VerCoche(coche);
               END;
               coche:=coche^.sig;
         END;
         letra:=CHR(ORD(letra)+1);
    END;  (*while*)
    IF vueltas>=100 THEN
       IO.WrStr("                           Felicidades campeon !! ");
    END;  (*if*)
  END;(*if principal*)
END Ganador;


(* Este proc. cambia la velocidad media de un coche.
   Devuelve error si no existe
   --no existira un coche si la pareja escuderia-numero no existe *)
PROCEDURE CambiarVelocidad(VAR rally:RALLY; escuderia:ARRAY OF CHAR;
                               numero,velocidad:INTEGER;VAR bien:BOOLEAN);
VAR
  l:LISTA;
  vacio:BOOLEAN;
BEGIN
  bien:=TRUE;
  vacio:=EsVacio(rally,bien);
  IF NOT(bien) THEN                 (*error *)
  ELSIF vacio  THEN                 (*EstaVacio *)
     bien:=FALSE
  ELSIF (NOT(EstaEsc(rally^[CAP(escuderia[0])],escuderia))) OR
        (NOT(EstaNum(rally,numero))) THEN
    bien:=FALSE                     (*El coche no esta *)
  ELSE
    l:=rally^[CAP(escuderia[0])];
    WHILE (MiStr.Comparar(escuderia,l^.nombre)<>0) DO
       l:=l^.sig;
    END;
    IF numero<>l^.numero THEN
       bien:=FALSE                  (*El coche no esta *)
    ELSE
       Calcular(l);                 (*Calcula el numero de vueltas
                                      y los metros recorridos hasta ahora
                                      con la velocidad que tenia y a partir
                                      de la ultima vez que se midio
                                      la distancia recorrida por este coche*)
       l^.velocidad:=velocidad;
    END
  END (*iff principal*)
END CambiarVelocidad;


(* Presenta de forma ordenada todos los coches que hay en carrera
   sin ofrecer los datos vacios y con pausas para mas comodidad. *)

PROCEDURE Visualizar(rally:RALLY;VAR bien:BOOLEAN);
VAR
  p, s:CHAR;
  l:LISTA;
  i:CARDINAL;
  vacio:BOOLEAN;
BEGIN
  bien:=TRUE;
  vacio:=EsVacio(rally,bien);
  IF vacio THEN
     bien:=FALSE
  ELSIF bien THEN         (* No esta vacion y no hubo errores *)
     Cronometrar(rally);
     p:='A';
     i:=0;
     WHILE p<='Z' DO
          l:=rally^[p];
          WHILE l<>NIL DO
            IF i<4 THEN     (* como maximo visualiza 4 coches seguidos
                               y continua cuando pulsamos 'S' *)
               VerCoche(l);
               l:=l^.sig;
               INC(i);
            ELSE
               i:=0;
               IO.WrStr("     ...PULSE cualquier tecla para continuar");
               REPEAT UNTIL IO.KeyPressed();
             END; (*if*)
          END;
          p:=CHR(ORD(p)+1);
     END;
  END;
  (*if principal*)
END Visualizar;


(* Este proc. permite al usuario teclear un valor positivo que se interpretara
   como un incremento de la velocidad de todos los coches cuya
   escuderia empieze por una letra entre la 'a' y la 'm' y cuyo
   numero tenga mas de 2 digitos.  *)
PROCEDURE Acelerar(VAR rally:RALLY; aceleracion:INTEGER;VAR bien:BOOLEAN);
VAR
  p:CHAR;
  l:LISTA;
  v:INTEGER;
  vacio:BOOLEAN;
BEGIN
  bien:=TRUE;
  vacio:=EsVacio(rally,bien);
  IF vacio THEN
    bien:=FALSE        (* No puedo acelerar nada *)
  ELSIF bien THEN
    p:='A';
    WHILE p<='M' DO
      l:=rally^[p];
      WHILE l<>NIL DO
        IF l^.numero>99 THEN
          v:=l^.velocidad+aceleracion;
          CambiarVelocidad(rally,l^.nombre,l^.numero,v,bien);
        END;
        l:=l^.sig;
      END;
      p:=CHR(ORD(p)+1);
    END;
  END;
END Acelerar;

PROCEDURE Destruir(VAR rally:RALLY;VAR bien:BOOLEAN);
VAR
  p:CHAR;
  l, b:LISTA;
  vacio:BOOLEAN;
BEGIN
  bien:=TRUE;
  vacio:=EsVacio(rally,bien);
  IF NOT(bien) THEN     (* Simplemente devuelvo bien a FALSE *)
  ELSIF vacio THEN
     DISPOSE(rally);
  ELSE
     p:='A';
     WHILE p<='Z' DO
       l:=rally^[p];
       WHILE l<>NIL DO
          b:=l;
          l:=l^.sig;
          DISPOSE(b);
       END;
       p:=CHR(ORD(p)+1);
     END;
     DISPOSE(rally);   (* Muy importante destruir la estructura *)
  END;
END Destruir;

PROCEDURE Volcado(rally:RALLY; nomfich:MiStr.TCadena; VAR bien:BOOLEAN);
VAR
  fichero:FIO.File;
  aux:LISTA;
  p:CHAR;

BEGIN
    bien:=TRUE;
    IF (rally=NIL) OR (EsVacio(rally,bien)) THEN
       bien:=FALSE;
    ELSE
       p:='A';
       fichero:=FIO.Create(nomfich);
       WHILE (p<='Z') DO
             aux:=rally^[p];
             WHILE aux<>NIL DO
                   FIO.WrStr(fichero,aux^.nombre);
                   FIO.WrLn(fichero);
                   FIO.WrInt(fichero,aux^.numero,0);
                   FIO.WrChar(fichero,' ');
                   FIO.WrInt(fichero,aux^.velocidad,0);
                   FIO.WrChar(fichero,' ');
                   FIO.WrLngCard(fichero,aux^.crono,0);
                   FIO.WrChar(fichero,' ');
                   FIO.WrCard(fichero,aux^.vueltas,0);
                   FIO.WrChar(fichero,' ');
                   FIO.WrCard(fichero,aux^.metros,0);
                   FIO.WrChar(fichero,' ');
                   aux:=aux^.sig;
             END;
             INC(p);
         END;
         FIO.Close(fichero);
    END;
END Volcado;

PROCEDURE Revolcado(VAR rally:RALLY; nomfich:MiStr.TCadena; VAR bien:BOOLEAN);
VAR
  fichero:FIO.File;
  nuevo:LISTA;
  nombre:MiStr.TCadena;

BEGIN
    bien:=TRUE;
    IF (rally=NIL) OR NOT FIO.Exists(nomfich) THEN
       bien:=FALSE;
    ELSE
       fichero:=FIO.Open(nomfich);
       FIO.RdStr(fichero, nombre);
       WHILE NOT FIO.EOF AND bien DO
             IF NOT Available(SIZE(NODO)) THEN
                bien:=FALSE;
             ELSE
                NEW(nuevo);
                MiStr.Copiar(nuevo^.nombre,nombre);
                nuevo^.numero:=FIO.RdInt(fichero);
                nuevo^.velocidad:=FIO.RdInt(fichero);
                nuevo^.crono:=FIO.RdLngCard(fichero);
                nuevo^.vueltas:=FIO.RdCard(fichero);
                nuevo^.metros:=FIO.RdCard(fichero);
                nuevo^.sig:=NIL;
                Colocar(rally,nuevo);
                FIO.RdStr(fichero,nombre);
             END;
         END;
         FIO.Close(fichero);
    END;
END Revolcado;





END Rally.